@@ -21,6 +21,7 @@
 #include <config.h>
 #include <gio/gio.h>
 #include <glib-object.h>
+#include <glib/gi18n-lib.h>
 #include <gtk/gtk.h>
 #include <string.h>
 #include "gtkfilesystemgio.h"
@@ -88,6 +89,7 @@ struct GtkFileSystemHandleGio
   guint source_id;
   gpointer callback;
   gpointer data;
+  guint tried_mount : 1;
 };
 
 struct GtkFileFolderGioClass
@@ -196,6 +198,9 @@ static gchar *               gtk_file_system_gio_get_bookmark_label (GtkFileSyst
 static void                  gtk_file_system_gio_set_bookmark_label (GtkFileSystem                *file_system,
 								     const GtkFilePath            *path,
 								     const gchar                  *label);
+static void                  query_info_callback                    (GObject      *source_object,
+								     GAsyncResult *result,
+								     gpointer      user_data);
 
 /* GtkFileFolderGio methods */
 static GType gtk_file_folder_gio_get_type       (void);
@@ -211,6 +216,14 @@ static gboolean              gtk_file_folder_gio_list_children    (GtkFileFolder
 								   GError                        **error);
 static gboolean              gtk_file_folder_gio_is_finished_loading (GtkFileFolder               *folder);
 
+/* The pointers we return for a GtkFileSystemVolume are opaque tokens; they are
+ * really pointers to GDrive, GVolume or GMount objects.  We need an extra
+ * token for the fake "File System" volume.  So, we'll return the pointer to
+ * this particular string.
+ */
+static const char *root_volume_token = "File System";
+#define IS_ROOT_VOLUME_TOKEN(volume) ((gpointer) (volume) == (gpointer) root_volume_token)
+
 /* GtkFileSystem module methods */
 void                         fs_module_init     (GTypeModule    *module);
 void                         fs_module_exit     (void);
@@ -263,6 +276,8 @@ get_volumes_list (GtkFileSystemGio *file_system)
 
   if (file_system->volumes)
     {
+      /* Remove the first one by hand */
+      file_system->volumes = g_slist_remove (file_system->volumes, file_system->volumes->data);
       g_slist_foreach (file_system->volumes, (GFunc) g_object_unref, NULL);
       g_slist_free (file_system->volumes);
       file_system->volumes = NULL;
@@ -368,6 +383,9 @@ get_volumes_list (GtkFileSystemGio *file_system)
       g_object_unref (mount);
     }
   g_list_free (mounts);
+
+  /* And finally, add the root volume to the front of the list */
+  file_system->volumes = g_slist_prepend (file_system->volumes, (gpointer) root_volume_token);
 }
 
 static void
@@ -378,7 +396,9 @@ volumes_drives_changed (GVolumeMonitor *volume_monitor,
   GtkFileSystemGio *impl;
 
   impl = GTK_FILE_SYSTEM_GIO (user_data);
+  gdk_threads_enter ();
   g_signal_emit_by_name (impl, "volumes-changed");
+  gdk_threads_leave ();
 }
 
 static gchar *
@@ -448,6 +468,8 @@ gtk_file_system_gio_dispose (GObject *object)
 
   if (impl->volumes)
     {
+      /* Remove the first one by hand */
+      impl->volumes = g_slist_remove (impl->volumes, impl->volumes->data);
       g_slist_foreach (impl->volumes, (GFunc) g_object_unref, NULL);
       g_slist_free (impl->volumes);
       impl->volumes = NULL;
@@ -555,7 +577,8 @@ gtk_file_system_gio_list_volumes (GtkFileSystem *file_system)
 
   file_system_gio = GTK_FILE_SYSTEM_GIO (file_system);
   list = g_slist_copy (file_system_gio->volumes);
-  g_slist_foreach (list, (GFunc) g_object_ref, NULL);
+  /* Don't try to ref the "File System" one */
+  g_slist_foreach (list->next, (GFunc) g_object_ref, NULL);
 
   return list;
 }
@@ -568,15 +591,21 @@ gtk_file_system_gio_get_volume_for_path (GtkFileSystem     *file_system,
   GFile *file;
   GMount *mount;
   GSList *list;
+  const char *uri;
 
   DEBUG ("get_volume_for_path");
 
+  uri = gtk_file_path_get_string (path);
+  if (strcmp (uri, "file:///") == 0)
+    return (GtkFileSystemVolume *) root_volume_token;
+
   file_system_gio = GTK_FILE_SYSTEM_GIO (file_system);
-  file = get_file_from_path (path);
+  file = g_file_new_for_uri (uri);
 
   g_return_val_if_fail (file != NULL, NULL);
 
-  for (list = file_system_gio->volumes; list; list = list->next)
+  /* Skip the first item on the list! */
+  for (list = file_system_gio->volumes->next; list; list = list->next)
     {
       if (g_type_is_a (G_OBJECT_TYPE (list->data), G_TYPE_MOUNT))
         {
@@ -627,7 +656,9 @@ enumerator_files_callback (GObject      *source_object,
 				     NULL, NULL, NULL);
 
       folder->finished_loading = TRUE;
+      gdk_threads_enter ();
       g_signal_emit_by_name (folder, "finished-loading", 0);
+      gdk_threads_leave ();
       g_object_unref (folder);
       return;
     }
@@ -651,7 +682,9 @@ enumerator_files_callback (GObject      *source_object,
 				      enumerator_files_callback,
 				      folder);
 
+  gdk_threads_enter ();
   g_signal_emit_by_name (folder, "files-added", added_files);
+  gdk_threads_leave ();
   g_slist_foreach (added_files, (GFunc) g_free, NULL);
   g_slist_free (added_files);
 
@@ -674,10 +707,14 @@ directory_monitor_changed (GFileMonitor      *monitor,
   switch (event)
     {
     case G_FILE_MONITOR_EVENT_CREATED:
+      gdk_threads_enter ();
       g_signal_emit_by_name (folder, "files-added", files);
+      gdk_threads_leave ();
       break;
     case G_FILE_MONITOR_EVENT_DELETED:
+      gdk_threads_enter ();
       g_signal_emit_by_name (folder, "files-removed", files);
+      gdk_threads_leave ();
       break;
     default:
       break;
@@ -727,9 +764,11 @@ enumerate_children_callback (GObject      *source_object,
       g_object_unref (enumerator);
     }
 
+  gdk_threads_enter ();
   ((GtkFileSystemGetFolderCallback) handle->callback) (GTK_FILE_SYSTEM_HANDLE (handle),
 						       GTK_FILE_FOLDER (folder),
 						       error, handle->data);
+  gdk_threads_leave ();
 }
 
 static GtkFileSystemHandle *
@@ -763,6 +802,41 @@ gtk_file_system_gio_get_folder (GtkFileSystem                  *file_system,
   return GTK_FILE_SYSTEM_HANDLE (handle);
 }
 
+static gchar *
+get_icon_for_special_directory (const char *uri)
+{
+  const char *special_dir;
+  char *special_uri;
+
+  if (!uri)
+    return NULL;
+
+  if (strcmp (uri, "file:///") == 0)
+    return "gnome-dev-harddisk";
+
+  special_dir = g_get_user_special_dir (G_USER_DIRECTORY_DESKTOP);
+  special_uri = g_filename_to_uri (special_dir, NULL, NULL);
+  if (strcmp (uri, special_uri) == 0)
+    {
+      g_free (special_uri);
+      return "gnome-fs-desktop";
+    }
+
+  g_free (special_uri);
+
+  special_dir = g_get_home_dir ();
+  special_uri = g_filename_to_uri (special_dir, NULL, NULL);
+  if (strcmp (uri, special_uri) == 0)
+    {
+      g_free (special_uri);
+      return "gnome-fs-home";
+    }
+
+  g_free (special_uri);
+
+  return NULL;
+}
+
 static gchar *
 get_icon_string (GIcon *icon)
 {
@@ -787,14 +861,14 @@ get_icon_string (GIcon *icon)
 
       icon_file = g_file_icon_get_file (G_FILE_ICON (icon));
       name = g_file_get_path (icon_file);
-      g_object_unref (icon_file);
     }
 
   return name;
 }
 
 static GtkFileInfo *
-translate_file_info (GFileInfo *file_info)
+translate_file_info (const char *uri,
+		     GFileInfo *file_info)
 {
   GtkFileInfo *info;
   gboolean is_folder;
@@ -818,19 +892,58 @@ translate_file_info (GFileInfo *file_info)
     gtk_file_info_set_icon_name (info, thumbnail_path);
   else
     {
-      GIcon *icon;
-      gchar *name;
+      const gchar *icon_name;
 
-      icon = g_file_info_get_icon (file_info);
-      name = get_icon_string (icon);
-      gtk_file_info_set_icon_name (info, name);
+      icon_name = get_icon_for_special_directory (uri);
+      if (icon_name)
+        {
+          gtk_file_info_set_icon_name (info, icon_name);
+	}
+      else
+        {
+          GIcon *icon;
+          gchar *name;
 
-      g_free (name);
+	  icon = g_file_info_get_icon (file_info);
+	  name = get_icon_string (icon);
+	  gtk_file_info_set_icon_name (info, name);
+
+	  g_free (name);
+	}
     }
 
   return info;
 }
 
+static void
+mount_async_callback (GObject      *source_object,
+		      GAsyncResult *result,
+		      gpointer      user_data)
+{
+  GtkFileSystemHandleGio *handle;
+  GError *error = NULL;
+  GFile *file;
+
+  DEBUG ("mount_async_callback");
+
+  file = G_FILE (source_object);
+  handle = GTK_FILE_SYSTEM_HANDLE_GIO (user_data);
+  if (g_file_mount_enclosing_volume_finish (file, result, &error))
+    {
+      g_file_query_info_async (file, "standard,time,thumbnail::*", 0, 0,
+			       handle->cancellable,
+			       query_info_callback,
+			       handle);
+    }
+  else
+    {
+      gdk_threads_enter ();
+      ((GtkFileSystemGetInfoCallback) handle->callback) (GTK_FILE_SYSTEM_HANDLE (handle),
+							 NULL, error, handle->data);
+      gdk_threads_leave ();
+    }
+}
+
 static void
 query_info_callback (GObject      *source_object,
 		     GAsyncResult *result,
@@ -850,12 +963,29 @@ query_info_callback (GObject      *source_object,
 
   if (file_info)
     {
-      info = translate_file_info (file_info);
+      char *uri;
+
+      uri = g_file_get_uri (file);
+      info = translate_file_info (uri, file_info);
+      g_free (uri);
       g_object_unref (file_info);
     }
+  else if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_MOUNTED) && !handle->tried_mount)
+    {
+      /* If it's not mounted, try to mount it ourselves */
+      g_error_free (error);
+      handle->tried_mount = TRUE;
+      g_file_mount_enclosing_volume (file, G_MOUNT_MOUNT_NONE, NULL,
+				     handle->cancellable,
+				     mount_async_callback,
+				     handle);
+      return;
+    }
 
+  gdk_threads_enter ();
   ((GtkFileSystemGetInfoCallback) handle->callback) (GTK_FILE_SYSTEM_HANDLE (handle),
 						     info, error, handle->data);
+  gdk_threads_leave ();
 
   if (info)
     gtk_file_info_free (info);
@@ -882,6 +1012,7 @@ gtk_file_system_gio_get_info (GtkFileSystem                *file_system,
   handle->cancellable = g_cancellable_new ();
   handle->callback = callback;
   handle->data = data;
+  handle->tried_mount = FALSE;
 
   g_file_query_info_async (file, "standard,time,thumbnail::*", 0, 0,
 			   handle->cancellable,
@@ -913,8 +1044,11 @@ create_folder_callback (gpointer data)
 
   g_file_make_directory (file, handle->cancellable, &error);
 
+  gdk_threads_enter ();
   ((GtkFileSystemCreateFolderCallback) handle->callback) (GTK_FILE_SYSTEM_HANDLE (handle),
 							  idle_data->path, error, handle->data);
+  gdk_threads_leave ();
+
   g_object_unref (file);
   gtk_file_path_free (idle_data->path);
   g_slice_free (CreateFolderData, idle_data);
@@ -979,6 +1113,9 @@ gtk_file_system_gio_volume_free (GtkFileSystem       *file_system,
 				 GtkFileSystemVolume *volume)
 {
   DEBUG ("volume_free");
+  if (IS_ROOT_VOLUME_TOKEN(volume))
+    return;
+
   g_object_unref (G_OBJECT (volume));
 }
 
@@ -991,6 +1128,9 @@ gtk_file_system_gio_volume_get_base_path (GtkFileSystem       *file_system,
 
   DEBUG ("volume_get_base_path");
 
+  if (IS_ROOT_VOLUME_TOKEN (file_system_volume))
+    return gtk_file_path_new_dup ("file:///");
+
   path = NULL;
 
   if (g_type_is_a (G_OBJECT_TYPE (file_system_volume), G_TYPE_MOUNT))
@@ -1027,6 +1167,9 @@ gtk_file_system_gio_volume_get_is_mounted (GtkFileSystem       *file_system,
 
   DEBUG ("volume_get_is_mounted");
 
+  if (IS_ROOT_VOLUME_TOKEN (file_system_volume))
+    return TRUE;
+
   mounted = FALSE;
 
   if (g_type_is_a (G_OBJECT_TYPE (file_system_volume), G_TYPE_MOUNT))
@@ -1060,9 +1203,12 @@ volume_mount_cb (GObject *source_object,
 
   g_volume_mount_finish (G_VOLUME (source_object), res, &error);
 
+  gdk_threads_enter ();
   ((GtkFileSystemVolumeMountCallback) handle->callback) (GTK_FILE_SYSTEM_HANDLE (handle),
                                                          (GtkFileSystemVolume *) source_object,
                                                          error, handle->data);
+  gdk_threads_leave ();
+
   if (error)
     g_error_free (error);
 }
@@ -1079,9 +1225,12 @@ drive_poll_for_media_cb (GObject *source_object,
 
   g_drive_poll_for_media_finish (G_DRIVE (source_object), res, &error);
 
+  gdk_threads_enter ();
   ((GtkFileSystemVolumeMountCallback) handle->callback) (GTK_FILE_SYSTEM_HANDLE (handle),
                                                          (GtkFileSystemVolume *) source_object,
                                                          error, handle->data);
+  gdk_threads_leave ();
+
   if (error)
     g_error_free (error);
 }
@@ -1133,6 +1282,9 @@ gtk_file_system_gio_volume_get_display_name (GtkFileSystem       *file_system,
 
   DEBUG ("volume_get_display_name");
 
+  if (IS_ROOT_VOLUME_TOKEN (file_system_volume))
+    return g_strdup (_("File System"));
+
   if (g_type_is_a (G_OBJECT_TYPE (file_system_volume), G_TYPE_DRIVE))
     {
       GDrive *drive = G_DRIVE (file_system_volume);
@@ -1162,6 +1314,9 @@ gtk_file_system_gio_volume_get_icon_name (GtkFileSystem        *file_system,
 
   DEBUG ("volume_get_icon_name");
 
+  if (IS_ROOT_VOLUME_TOKEN (file_system_volume))
+    return g_strdup ("gnome-dev-harddisk");
+
   if (g_type_is_a (G_OBJECT_TYPE (file_system_volume), G_TYPE_DRIVE))
     {
       GDrive *drive = G_DRIVE (file_system_volume);
@@ -1175,6 +1330,19 @@ gtk_file_system_gio_volume_get_icon_name (GtkFileSystem        *file_system,
   else if (g_type_is_a (G_OBJECT_TYPE (file_system_volume), G_TYPE_MOUNT))
     {
       GMount *mount = G_MOUNT (file_system_volume);
+      GFile *file;
+      const char *icon_name;
+      char *uri;
+
+      file = g_mount_get_root (mount);
+      uri = g_file_get_uri (file);
+      g_object_unref (file);
+
+      icon_name = get_icon_for_special_directory (uri);
+      g_free (uri);
+      if (icon_name)
+        return g_strdup (icon_name);
+
       icon = g_mount_get_icon (mount);
     }
 
@@ -1729,15 +1897,17 @@ gtk_file_folder_gio_get_info (GtkFileFolder      *folder,
 {
   GtkFileFolderGio *folder_gio;
   GFileInfo *file_info;
+  const char *uri;
 
   DEBUG ("folder_get_info");
 
   folder_gio = GTK_FILE_FOLDER_GIO (folder);
+  uri = gtk_file_path_get_string (path);
   file_info = g_hash_table_lookup (folder_gio->children,
-				   gtk_file_path_get_string (path));
+				   uri);
 
   if (file_info)
-    return translate_file_info (file_info);
+    return translate_file_info (uri, file_info);
 
   return NULL;
 }
